/* Easy Machine: assember.
   This is a derivative work based on Jitter, which is released under
   the GNU General Public License, version 3 or later. */

#include <config.h>
#include <xalloc.h>

#include <stdio.h>
#include <string.h>
#include <stdbool.h>

#include <easy-machine/easy-machine-machine-interpreter.h>
#include <easy-machine/easy-machine-instructions-machine-generated.h>
#include <easy-machine/easy-machine-assembler.h>


/* Fatal errors.
 * ************************************************************************** */

#define easy_machine_asm_fatal_at(the_line_no, ...)                 \
  do                                                                \
    {                                                               \
      /* Ignore any failures in printing a fatal error message. */  \
      if (r->file_name != NULL)                                     \
        printf ("%s", r->file_name);                                \
      printf (":%i: ", the_line_no);                                \
      printf (__VA_ARGS__);                                         \
      printf ("\n");                                                \
      exit (EXIT_FAILURE);                                          \
    }                                                               \
  while (false)

#define easy_machine_asm_fatal(...)                                 \
  do                                                                \
    {                                                               \
      /* Ignore any failures in printing a fatal error message. */  \
      if (r->file_name != NULL)                                     \
        printf ("%s", r->file_name);                                \
      printf (":%i: ", r->line_no);                                 \
      printf (__VA_ARGS__);                                         \
      printf ("\n");                                                \
      exit (EXIT_FAILURE);                                          \
    }                                                               \
  while (false)




/* Machine-generated file.
 * ************************************************************************** */

#include <easy-machine/easy-machine-machine-assembler-machine-generated.c>




/* Routines.
 * ************************************************************************** */

void
easy_machine_asm_initialize_routine (struct easy_machine_routine *r)
{
  r->file_name = NULL;
  r->line_no = -1;
  r->current_section = easy_machine_asm_section_code;
  r->current_instruction = NULL;
  r->appending_bytes = false;
  r->appending_words = false;
  int i;
  for (i = 0; i < easy_machine_asm_section_no; i ++)
    jitter_dynamic_buffer_initialize (r->section_contents + i);
  jitter_hash_initialize (& r->label_to_place);
}




/* Routine editing.
 * ************************************************************************** */

static void
easy_machine_asm_close_instruction_if_any (struct easy_machine_routine *r)
{
  r->current_instruction = NULL;
  r->appending_bytes = false;
  r->appending_words = false;
}

static struct easy_machine_asm_element *
easy_machine_asm_open_element (struct easy_machine_routine *r)
{
  easy_machine_asm_close_instruction_if_any (r);
  struct jitter_dynamic_buffer *db
    = r->section_contents + r->current_section;
  struct easy_machine_asm_element *new_element
    = jitter_dynamic_buffer_reserve (db, sizeof (struct easy_machine_asm_element));
  new_element->line_no = r->line_no;
  return new_element;
}

void
easy_machine_asm_enter_section (struct easy_machine_routine *r,
                                enum easy_machine_asm_section section)
{
  easy_machine_asm_close_instruction_if_any (r);
  r->current_section = section;
}

void
easy_machine_asm_append_words (struct easy_machine_routine *r)
{
  easy_machine_asm_close_instruction_if_any (r);
  r->appending_words = true;
}

void
easy_machine_asm_append_bytes (struct easy_machine_routine *r)
{
  easy_machine_asm_close_instruction_if_any (r);
  r->appending_bytes = true;
}

void
easy_machine_asm_append_datum_symbolic_label (struct easy_machine_routine *r,
                                              const char *name)
{
  if (r->appending_bytes)
    easy_machine_asm_fatal ("cannot append label %s to bytes (use .word instead)", name);
  struct easy_machine_asm_element *new_element
    = easy_machine_asm_open_element (r);
  new_element->kind = easy_machine_asm_element_kind_datum_label;
  new_element->label = easy_machine_string_clone (name);
}

void
easy_machine_asm_append_datum_literal (struct easy_machine_routine *r,
                                       easy_machine_word w)
{
  if (r->appending_bytes && (w < 0 || w > 255))
    easy_machine_asm_fatal ("cannot append %li as a byte (did you mean to use .word instead?)", (long) w);
  bool appending_bytes = r->appending_bytes;
  struct easy_machine_asm_element *new_element
    = easy_machine_asm_open_element (r);
  if (appending_bytes)
    {
      new_element->kind = easy_machine_asm_element_kind_datum_byte;
      new_element->byte = w;
      r->appending_bytes = true; // keep for the following
    }
  else
    {
      new_element->kind = easy_machine_asm_element_kind_datum_word;
      new_element->word = w;
      r->appending_words = true; // keep for the following
    }
}

static void
easy_machine_asm_open_instruction (struct easy_machine_routine *r,
                                   const char *mnemonic)
{
  struct easy_machine_asm_element *new_element
    = easy_machine_asm_open_element (r);
  new_element->kind = easy_machine_asm_element_kind_instruction;
  struct easy_machine_asm_instruction *new_instruction
    = & new_element->instruction;
  new_instruction->line_no = new_element->line_no;
  new_instruction->mnemonic = easy_machine_string_clone (mnemonic);
  new_instruction->arg_no = 0;
  r->current_instruction = new_instruction;
}

void
easy_machine_asm_append_symbolic_label (struct easy_machine_routine *r,
                                        const char *name)
{
  struct easy_machine_asm_element *new_element
    = easy_machine_asm_open_element (r);
  new_element->kind = easy_machine_asm_element_kind_label;
  new_element->label = easy_machine_string_clone (name);
}

static struct easy_machine_asm_instruction_argument *
easy_machine_asm_append_instruction_parameter (struct easy_machine_routine *r)
{
  if (r->current_instruction == NULL)
    easy_machine_asm_fatal ("appending instruction parameter to non-existing instruction: this should never happen");
  struct easy_machine_asm_instruction *i = r->current_instruction;
  i->arg_no ++;
  if (i->arg_no > EASY_MACHINE_MAX_ARGS)
    easy_machine_asm_fatal ("too many instruction arguments");
  else
    return r->current_instruction->args + i->arg_no - 1;
}

void
easy_machine_asm_append_instruction_name (struct easy_machine_routine *r,
                                          const char *name)
{
  easy_machine_asm_open_instruction (r, name);
}

void
easy_machine_asm_append_literal_parameter (struct easy_machine_routine *r,
                                           easy_machine_word w)
{
  struct easy_machine_asm_instruction_argument *new_arg
    = easy_machine_asm_append_instruction_parameter (r);
  new_arg->kind = easy_machine_asm_instruction_argument_kind_numeric;
  new_arg->numeric_literal = w;
}

void
easy_machine_asm_append_register_parameter (struct easy_machine_routine *r,
                                            easy_machine_register_index idx)
{
  struct easy_machine_asm_instruction_argument *new_arg
    = easy_machine_asm_append_instruction_parameter (r);
  if (idx < 0 || idx >= EASY_MACHINE_REGISTER_NO)
    easy_machine_asm_fatal ("register %%r%i invalid: the machine has %i registers numbered from %%r0 to %%r%i included", (int) idx, (int) EASY_MACHINE_REGISTER_NO, (int) (EASY_MACHINE_REGISTER_NO - 1));
  new_arg->kind = easy_machine_asm_instruction_argument_kind_register;
  new_arg->register_idx = idx;
}

void
easy_machine_asm_append_symbolic_label_parameter (struct easy_machine_routine *r,
                                                  const char *name)
{
  struct easy_machine_asm_instruction_argument *new_arg
    = easy_machine_asm_append_instruction_parameter (r);
  new_arg->kind = easy_machine_asm_instruction_argument_kind_label;
  new_arg->label_literal = easy_machine_string_clone (name);
}




/* Resolution.
 * ************************************************************************** */

easy_machine_word
easy_machine_section_load_addresses [easy_machine_asm_section_no];

const char *
easy_machine_asm_section_names [easy_machine_asm_section_no]
= {"code", "readonly", "readwrite"};

static void
easy_machine_asm_compute_section_sizes (struct easy_machine_routine *r)
{
  int si, ei;
  for (si = 0; si < easy_machine_asm_section_no; si ++)
    {
      r->section_sizes [si] = 0;
      size_t used_size
        = jitter_dynamic_buffer_size (r->section_contents + si);
      size_t element_no = used_size / sizeof (struct easy_machine_asm_element);
      struct easy_machine_asm_element *elements
        = jitter_dynamic_buffer_to_pointer (r->section_contents + si);
      for (ei = 0; ei < element_no; ei ++)
        {
          switch (elements [ei].kind)
            {
            case easy_machine_asm_element_kind_instruction:
              elements [ei].size = sizeof (struct easy_machine_instruction);
              break;
            case easy_machine_asm_element_kind_datum_byte:
              elements [ei].size = 1;
              break;
            case easy_machine_asm_element_kind_datum_word:
            case easy_machine_asm_element_kind_datum_label:
              elements [ei].size = sizeof (easy_machine_word);
              break;
            case easy_machine_asm_element_kind_label:
              elements [ei].size = 0;
              break;
            default:
              easy_machine_asm_fatal ("C impossible");
            }
          elements [ei].offset_from_section = r->section_sizes [si];
          r->section_sizes [si] += elements [ei].size;
        }
      r->section_sizes_rounded_up [si]
        = JITTER_NEXT_MULTIPLE_OF_POWER_OF_TWO (r->section_sizes [si],
                                                EASY_MACHINE_PAGE_SIZE);
    }
}

static void
easy_machine_asm_compute_load_addresses (struct easy_machine_routine *r,
                                         struct easy_machine_assmbler_command_line *cl)
{
  easy_machine_word next_free_page = EASY_MACHINE_MINIMUM_ADDRESS;
  int si;
  for (si = 0; si < easy_machine_asm_section_no; si ++)
    {
      easy_machine_section_load_addresses [si] = next_free_page;
      next_free_page += r->section_sizes_rounded_up [si];
    }
  for (si = 0; si < easy_machine_asm_section_no; si ++)
    if (cl->debug)
      printf ("Section \"%s\" loaded at 0x%lx, size %liB (rounded to %liB, %li pages)\n", easy_machine_asm_section_names [si], (unsigned long) easy_machine_section_load_addresses [si], (long) r->section_sizes [si], (long) r->section_sizes_rounded_up [si], (long) r->section_sizes_rounded_up [si] / EASY_MACHINE_PAGE_SIZE);
}

static easy_machine_word
easy_machine_resolve_label (struct easy_machine_routine *r,
                            const char *label,
                            int line_no)
{
  if (! jitter_string_hash_table_has (& r->label_to_place, label))
    easy_machine_asm_fatal_at (line_no, "unbound label %s", label);

  const union jitter_word w
    = jitter_string_hash_table_get (& r->label_to_place, label);
  return w.ufixnum;
}

static void
easy_machine_asm_resolve_labels (struct easy_machine_routine *r,
                                 struct easy_machine_assmbler_command_line *cl)
{
  /* First pass: fill a map from symbolic label to address. */
  int si, ei;
  for (si = 0; si < easy_machine_asm_section_no; si ++)
    {
      size_t used_size
        = jitter_dynamic_buffer_size (r->section_contents + si);
      size_t element_no = used_size / sizeof (struct easy_machine_asm_element);
      struct easy_machine_asm_element *elements
        = jitter_dynamic_buffer_to_pointer (r->section_contents + si);
      for (ei = 0; ei < element_no; ei ++)
        if (elements [ei].kind == easy_machine_asm_element_kind_label)
          {
            struct easy_machine_asm_element *element = elements + ei;
            if (jitter_string_hash_table_has (& r->label_to_place,
                                              element->label))
              easy_machine_asm_fatal_at (element->line_no, "label %s defined for the second time", element->label);
            easy_machine_word virtual_address
              = (easy_machine_section_load_addresses [si]
                 + element->offset_from_section);
            union jitter_word w = {.ufixnum = virtual_address};
            if (cl->debug)
              printf ("Label: %s -> 0x%lx\n", element->label, (unsigned long) virtual_address);
            jitter_string_hash_table_add (& r->label_to_place,
                                          element->label,
                                          w);
          }
    }
  /* Second pass: use the map to replace labels with literals. */
  for (si = 0; si < easy_machine_asm_section_no; si ++)
    {
      size_t used_size
        = jitter_dynamic_buffer_size (r->section_contents + si);
      size_t element_no = used_size / sizeof (struct easy_machine_asm_element);
      struct easy_machine_asm_element *elements
        = jitter_dynamic_buffer_to_pointer (r->section_contents + si);
      for (ei = 0; ei < element_no; ei ++)
        {
          struct easy_machine_asm_element *element = elements + ei;
          switch (element->kind)
            {
            case easy_machine_asm_element_kind_label:
            case easy_machine_asm_element_kind_datum_byte:
            case easy_machine_asm_element_kind_datum_word:
              /* Do nothing. */
              break;
            case easy_machine_asm_element_kind_instruction:
              {
                struct easy_machine_asm_instruction *i = & element->instruction;
                int ai;
                for (ai = 0; ai < i->arg_no; ai ++)
                  if (i->args [ai].kind
                      == easy_machine_asm_instruction_argument_kind_label)
                    {
                      i->args [ai].kind
                        = easy_machine_asm_instruction_argument_kind_numeric;
                      i->args [ai].numeric_literal
                        = easy_machine_resolve_label (r, i->args [ai].label_literal,
                                                      i->line_no);
                    }
                break;
              }
            case easy_machine_asm_element_kind_datum_label:
              element->kind = easy_machine_asm_element_kind_datum_word;
              element->word = easy_machine_resolve_label (r, element->label,
                                                          element->line_no);
              break;
            default:
              easy_machine_asm_fatal ("D impossible");
            }
        }
    }
}

static void
easy_machine_asm_resolve_opcodes (struct easy_machine_routine *r)
{
  int si, ei;
  for (si = 0; si < easy_machine_asm_section_no; si ++)
    {
      size_t used_size
        = jitter_dynamic_buffer_size (r->section_contents + si);
      size_t element_no = used_size / sizeof (struct easy_machine_asm_element);
      struct easy_machine_asm_element *elements
        = jitter_dynamic_buffer_to_pointer (r->section_contents + si);
      for (ei = 0; ei < element_no; ei ++)
        if (elements [ei].kind == easy_machine_asm_element_kind_instruction)
          easy_machine_asm_instruction_set_opcode (r, & elements [ei].instruction);
    }
}

static easy_machine_word
easy_machine_asm_argument_to_word (struct easy_machine_routine *r,
                                   const struct easy_machine_asm_instruction_argument *a,
                                   int line_no)
{
  switch (a->kind)
    {
    case easy_machine_asm_instruction_argument_kind_numeric:
      return a->numeric_literal;
    case easy_machine_asm_instruction_argument_kind_register:
      return a->register_idx;
    case easy_machine_asm_instruction_argument_kind_label:
      easy_machine_asm_fatal_at (line_no, "impossible: unresolved label after resolving");
    default:
      easy_machine_asm_fatal ("H impossible");
    }
}

static easy_machine_word
easy_machine_last_loaded_address = -1;

static void
easy_machine_asm_load_resolved (struct easy_machine_routine *r,
                                struct easy_machine_assmbler_command_line *cl)
{
  /* First pass: fill a map from symbolic label to address. */
  int si, ei;
  for (si = 0; si < easy_machine_asm_section_no; si ++)
    {
      size_t used_size
        = jitter_dynamic_buffer_size (r->section_contents + si);
      size_t element_no = used_size / sizeof (struct easy_machine_asm_element);
      struct easy_machine_asm_element *elements
        = jitter_dynamic_buffer_to_pointer (r->section_contents + si);
      for (ei = 0; ei < element_no; ei ++)
        {
          struct easy_machine_asm_element *element = elements + ei;
          easy_machine_word virtual_address
            = (easy_machine_section_load_addresses [si]
               + element->offset_from_section);
          easy_machine_last_loaded_address = virtual_address;
          switch (element->kind)
            {
            case easy_machine_asm_element_kind_label:
              /* Do nothing. */
              break;
            case easy_machine_asm_element_kind_datum_byte:
              easy_machine_store_byte (virtual_address, element->byte);
              break;
            case easy_machine_asm_element_kind_datum_word:
              easy_machine_store_word (virtual_address, element->word);
              break;
            case easy_machine_asm_element_kind_instruction:
              {
                if (cl->debug)
                  printf ("Instruction: %s at 0x%lx\n",
                          easy_machine_instruction_names [element->instruction.opcode],
                          (unsigned long) virtual_address);
                struct easy_machine_instruction i;
                i.opcode = element->instruction.opcode;
                i.arg0 = i.arg1 = i.arg2 = 0;
                if (element->instruction.arg_no > 3)
                  easy_machine_asm_fatal ("E impossible");
                if (element->instruction.arg_no > 2)
                  i.arg2 = easy_machine_asm_argument_to_word (r, element->instruction.args + 2, element->line_no);
                if (element->instruction.arg_no > 1)
                  i.arg1 = easy_machine_asm_argument_to_word (r, element->instruction.args + 1, element->line_no);
                if (element->instruction.arg_no > 0)
                  i.arg0 = easy_machine_asm_argument_to_word (r, element->instruction.args + 0, element->line_no);
                easy_machine_store_word (virtual_address + sizeof (easy_machine_word) * 0, i.opcode);
                easy_machine_store_word (virtual_address + sizeof (easy_machine_word) * 1, i.arg0);
                easy_machine_store_word (virtual_address + sizeof (easy_machine_word) * 2, i.arg1);
                easy_machine_store_word (virtual_address + sizeof (easy_machine_word) * 3, i.arg2);
                break;
              }
            case easy_machine_asm_element_kind_datum_label:
              easy_machine_asm_fatal ("F impossible");
            default:
              easy_machine_asm_fatal ("G impossible");
            }
        }
    }
}

static void
easy_machine_asm_dump (struct easy_machine_routine *r,
                       struct easy_machine_assmbler_command_line *cl)
{
  if (cl->dump_from_address == -1)
    cl->dump_from_address
      = easy_machine_section_load_addresses [0];
  if (cl->dump_to_address == -1)
    cl->dump_to_address
      = easy_machine_last_loaded_address + 1;

  if (cl->debug)
    printf ("Dumping %lu bytes [0x%lx, 0x%lx)...\n============================================\n",
            (unsigned long) (cl->dump_to_address - cl->dump_from_address),
            (unsigned long) cl->dump_from_address,
            (unsigned long) cl->dump_to_address);
  easy_machine_word i;
  for (i = cl->dump_from_address; i < cl->dump_to_address; i ++)
    putchar (easy_machine_load_byte (i));
  if (cl->debug)
    printf ("\n============================================\nDone dumping.\n");
}


void
easy_machine_asm_load (struct easy_machine_routine *r,
                       struct easy_machine_assmbler_command_line *cl)
{
  easy_machine_asm_close_instruction_if_any (r);

  easy_machine_asm_compute_section_sizes (r);
  easy_machine_asm_compute_load_addresses (r, cl);
  easy_machine_asm_resolve_labels (r, cl);
  easy_machine_asm_resolve_opcodes (r);

  easy_machine_asm_load_resolved (r, cl);

  if (cl->dump)
    easy_machine_asm_dump (r, cl);
}
